package Alt::CWB::ambs;

use warnings;
use strict;
our $VERSION = '2.2.102.6';

1;

__END__

=encoding UTF-8

=head1 NAME

Alt::CWB::ambs - A CPAN distribution of CWB

=head1 DESCRIPTION

The Open-CWB L<http://cwb.sourceforge.net/> project makes available
two modules to work with CWB version 3.0.x. Unfortunately they are not
available on CPAN, and therefore it is not possible to depend on them.

This B<ALTERNATE> distribution is exactly the same code as the one
prepared by Stefan Evert and company.

=head1 COPYRIGHT

Copyright (C) 1999-2010 Stefan Evert [http::/purl.org/stefan.evert]

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut
