# -*-cperl-*-
## Test whether all CWB modules can be loaded

use Test::More tests => 6;

BEGIN {
  use_ok('CWB');
  use_ok('CWB::Encoder');
  use_ok('CWB::CQP');
  use_ok('CWB::CEQL::Parser');
  use_ok('CWB::CEQL::String');
  use_ok('CWB::CEQL');
};
